table.maze {
    border-collapse: collapse;
}

td {
    text-align: center;
}

table.maze td.cell {
    position: relative;
    border: 1px solid black;
}

table.maze td.cell div.agent {
    position: absolute;
    left: 25%;
    top: 25%;;
    width: 50%;
    height: 50%;
    border-radius: 100%;
    background:rgba(69, 100, 186, 0.5);
}


td.wall {
    background: black;
}

td.terminal {
    background: rgb(246, 170, 246);
}

table.maze table td {
    width: .5rem;
    height: .5rem;
}



table.maze table td.arrow {
    color: transparent;
    font-weight: bold;
}

table.maze table td.value {
    position: relative;
    width: 2rem;
    height: 2rem;
}

